/* 
 * EDOMI beta (Version 1.59) - Intelligente Steuerung und Visualisierung KNX-basierter Elektro-Installationen.
 * @link http://www.edomi.de 
 * @author Dr. Christian Gärtner <dr.christian.gaertner@googlemail.com> 
 * @copyright 2015,2016,2017 Dr. Christian Gärtner 
 * @license Freeware 
 *
 *     This file and all other files are part of EDOMI.
 *
 *     EDOMI is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 *     Diese und alle anderen Dateien sind Teil von EDOMI.
 *
 *     EDOMI wird in der Hoffnung, dass es nützlich sein wird, aber
 *     OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 *     Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
*/ 

function class_camView() {
	var that=this;
	
	var src={};
	var dst={};
	var property={};
	var downscaled=false;
	var renderWorker=null;
	var renderWorkerBuffer=null;
	
	this.setProperty=function(p,value) {
		property[p]=value;
	}

	this.init=function() {	
		if (property.render) {
			src.canvas=document.createElement('canvas');
		} else {
			src.canvas=property.srccanvas;
		}
		if (!src.canvas) {return false;}
		src.ctx=src.canvas.getContext("2d");

		dst.canvas=property.dstcanvas;
		if (!dst.canvas) {return false;}
		dst.ctx=dst.canvas.getContext("2d");		

		that.loadImage();
	}
	
	this.loadImage=function() {	
		downscaled=false;
		src.img=new Image();
		src.img.onload=function() {
			that.render();	
		};
		src.img.onerror=function() {
			callback(false);
		};
		src.img.src=property.url;	
	}

	this.loadImageToDst=function() {	
		dst.canvas=property.dstcanvas;
		if (!dst.canvas) {return false;}
		dst.ctx=dst.canvas.getContext("2d");		

		var tmp=new Image();
		tmp.onload=function() {
			dst.canvas.width=tmp.width;
			dst.canvas.height=tmp.height;
			dst.ctx.drawImage(tmp,0,0,tmp.width,tmp.height);
			callback(true);
		};
		tmp.onerror=function() {
			callback(false);
		};
		tmp.src=property.url;	
	}

	this.render=function() {
		if (scaleSrcImage()) {
			getPropertys();
	
			if (property.render) {
				property.dstw=property.db_dstw;
				property.dsth=property.db_dsth;
			} else {
				if (property.aspect>=1) {
					property.dstw=parseInt(property.dstcanvassize);
					property.dsth=parseInt(property.dstcanvassize/property.aspect);
				} else {
					property.dstw=parseInt(property.dstcanvassize*property.aspect);
					property.dsth=parseInt(property.dstcanvassize);
				}
			}
	
			var res=getResolution();
			if (property.dstw>res[0] || property.dsth>res[1]) {
				property.dstw=res[0];
				property.dsth=res[1];
			}
	
			if (property.dstw<1) {property.dstw=1;}
			if (property.dsth<1) {property.dsth=1;}
	
			if (!property.render) {
				var psize=3;
				var size=psize*(Math.sqrt(src.w*src.w+src.h*src.h)/Math.sqrt(src.cw*src.cw+src.ch*src.ch));
				src.ctx.fillStyle=apps_colorSelected;
				for (var x=0;x<property.dstw;x++) {
					var p=camView_getPixel(property,x,0,property.dstw,property.dsth);
					if (p[0]>=0) {src.ctx.fillRect(p[0]-size/2,p[1]-size/2,size,size);}
					var p=camView_getPixel(property,x,property.dsth,property.dstw,property.dsth);
					if (p[0]>=0) {src.ctx.fillRect(p[0]-size/2,p[1]-size/2,size,size);}
				}
				for (var y=0;y<property.dsth;y++) {
					var p=camView_getPixel(property,0,y,property.dstw,property.dsth);
					if (p[0]>=0) {src.ctx.fillRect(p[0]-size/2,p[1]-size/2,size,size);}
					var p=camView_getPixel(property,property.dstw,y,property.dstw,property.dsth);
					if (p[0]>=0) {src.ctx.fillRect(p[0]-size/2,p[1]-size/2,size,size);}
				}
			}
			
			if (renderWorker) {
				renderWorkerBuffer=property;
			} else {
				render_worker();
			}
		}
	}
	
	function render_worker() {

/*
### im Prinzip ok, nutzt aber nicht den Worker-Prozess (wg. DOM-Zugriff) => bei sehr großen Auflösungen wird DOM evtl. etwas stocken
		//Sonderfall: 2D-Ausschnitt ohne Rotation (schneller)
		if (property.srctyp==1 && property.db_a2==0) {
			var x0=((1-0/property.dstw*2))*property.fov*property.aspect;
			var y0=((1-0/property.dsth*2))*property.fov;
			var x1=((1-property.dstw/property.dstw*2))*property.fov*property.aspect;
			var y1=((1-property.dsth/property.dsth*2))*property.fov;		
			var xr0=parseInt((x0+property.xoffset)*property.srcR+property.srcMx)*(100/property.db_srcs);
			var yr0=parseInt((y0+property.yoffset)*property.srcR+property.srcMy)*(100/property.db_srcs);
			var xr1=parseInt((x1+property.xoffset)*property.srcR+property.srcMx)*(100/property.db_srcs);
			var yr1=parseInt((y1+property.yoffset)*property.srcR+property.srcMy)*(100/property.db_srcs);
			if (xr1>=0 && yr1>=0 && xr0<src.w && yr0<src.h) {							
				if (renderWorker) {renderWorker.terminate();}
				renderWorker=null;

				if (!property.render) {
					if (property.aspect>=1) {
						dst.canvas.style.width=parseInt(property.dstcanvassize)+"px";
						dst.canvas.style.height=parseInt(property.dstcanvassize/property.aspect)+"px";			
					} else {
						dst.canvas.style.width=parseInt(property.dstcanvassize*property.aspect)+"px";
						dst.canvas.style.height=parseInt(property.dstcanvassize)+"px";			
					}
				}
				dst.canvas.width=property.dstw;
				dst.canvas.height=property.dsth;
				dst.ctx.drawImage(src.img,xr1,yr1,(xr0-xr1),(yr0-yr1),0,0,property.dstw,property.dsth);
		
				callback(true);
				return;
			}
		}
*/

		dst.imgdata=new ImageData(property.dstw,property.dsth);
		renderWorker=new Worker("../shared/js/camview_worker.js");
		renderWorker.onmessage=render_workerMsg;
		renderWorker.postMessage({
			dstw:property.dstw,
			dsth:property.dsth,
			aspect:property.aspect,
			srctyp:property.srctyp,
			fov:property.fov,
			angle1:property.angle1,
			angle2:property.angle2,
			xoffset:property.xoffset,
			yoffset:property.yoffset,
			distort:property.distort,
			srcw:property.srcw,
			srch:property.srch,
			srcMx:property.srcMx,
			srcMy:property.srcMy,
			srcR:property.srcR,
			sin_a_1:property.sin_a_1,
			cos_a_1:property.cos_a_1,
			sinZ:property.sinZ,
			cosZ:property.cosZ,
			srcdata:src.imgdata,
			dstdata:dst.imgdata
		});		
	}

	function render_workerMsg(event) {

		if (!property.render) {
			if (property.aspect>=1) {
				dst.canvas.style.width=parseInt(property.dstcanvassize)+"px";
				dst.canvas.style.height=parseInt(property.dstcanvassize/property.aspect)+"px";			
			} else {
				dst.canvas.style.width=parseInt(property.dstcanvassize*property.aspect)+"px";
				dst.canvas.style.height=parseInt(property.dstcanvassize)+"px";			
			}
		}

		dst.canvas.width=event.data.dstdata.width;
		dst.canvas.height=event.data.dstdata.height;
		dst.ctx.putImageData(event.data.dstdata,0,0);

		renderWorker.terminate();
		renderWorker=null;

		callback(true);

		if (renderWorkerBuffer) {
			property=renderWorkerBuffer;
			renderWorkerBuffer=null;
			render_worker();
		}
	};

	function scaleSrcImage() {
		if (downscaled!=property.db_srcs) {
			var w=parseInt(src.img.width*property.db_srcs/100);
			var h=parseInt(src.img.height*property.db_srcs/100);
			if (w==0 || h==0) {return false;}

			if (property.srctyp==1) {			//2D
				src.w=w;
				src.h=h;
				if (src.w>=src.h) {
					src.r=src.w/2;
				} else {
					src.r=src.h/2;
				}
			} else if (property.srctyp==2) {	//3D: Kreis
				if (w>=h) {
					src.w=w;
					src.h=h;
					src.r=src.h/2;
				} else {
					src.w=w;
					src.h=h;
					src.r=src.w/2;
				}
			} else if (property.srctyp==3) {	//3D: Ellipse
				if (w>=h) {
					src.w=w;
					src.h=w;
					src.r=src.h/2;
				} else {
					src.w=h;
					src.h=h;
					src.r=src.w/2;
				}
			}

			src.canvas.width=src.w;
			src.canvas.height=src.h;
			src.ctx.drawImage(src.img,0,0,src.w,src.h);
			src.rawimgdata=src.ctx.getImageData(0,0,src.w,src.h);

			if (!property.render) {
				var cw=property.srccanvassize;
				var ch=property.srccanvassize;
				var a=src.w/src.h;
				if (a<1) {cw*=a;}
				if (a>1) {ch/=a;}
				src.canvas.style.width=parseInt(cw)+"px";
				src.canvas.style.height=parseInt(ch)+"px";
				src.cw=parseInt(cw);
				src.ch=parseInt(ch);
			}
			downscaled=property.db_srcs;
		}
		src.ctx.putImageData(src.rawimgdata,0,0);
		src.imgdata=src.ctx.getImageData(0,0,src.w,src.h);
		return true;
	}
	
	function getPropertys() {	
		if (property.db_dstw<=0) {property.db_dstw=1;}
		if (property.db_dsth<=0) {property.db_dsth=1;}
		property.aspect=property.db_dstw/property.db_dsth;
		
		if (property.srctyp==1) {property.fov=1-property.db_zoom/500;} else if (property.srctyp>=2) {property.fov=((1-property.db_zoom/500)*90)/180*Math.PI;}
		if (property.fov==0) {property.fov=0.001;}	
		property.angle1=property.db_a1*Math.PI/180;
		property.angle2=property.db_a2*Math.PI/180;
		property.xoffset=property.db_x/100;
		property.yoffset=property.db_y/100;
		if (property.db_srcd>=0) {
			property.distort=Math.PI*(1+property.db_srcd/100);
		} else {
			property.distort=Math.PI/(1+Math.abs(property.db_srcd/100));
		}

		property.srcMx=src.w/2;
		property.srcMy=src.h/2;		
		property.srcR=parseInt(src.r*(property.db_srcr+1000)/1000);
		property.sin_a_1=Math.sin(property.angle1);
		property.cos_a_1=Math.cos(property.angle1);
		property.sinZ=Math.sin(property.angle2);
		property.cosZ=Math.cos(property.angle2);
		property.srcw=src.w;
		property.srch=src.h;
	}

	function getResolution() {
		var ax=new Array();
		var ay=new Array();
		
		var p1=camView_getPixel(property,0+0.001,property.db_dsth+0.001,property.db_dstw,property.db_dsth,true);
		var p2=camView_getPixel(property,property.db_dstw+0.001,property.db_dsth+0.001,property.db_dstw,property.db_dsth,true);
		ax.push(parseInt(Math.sqrt((p1[0]-p2[0])*(p1[0]-p2[0])+(p1[1]-p2[1])*(p1[1]-p2[1])+(p1[2]-p2[2])*(p1[2]-p2[2]))*property.srcR));
		var p1=camView_getPixel(property,0+0.001,property.db_dsth/2+0.001,property.db_dstw,property.db_dsth,true);
		var p2=camView_getPixel(property,property.db_dstw+0.001,property.db_dsth/2+0.001,property.db_dstw,property.db_dsth,true);
		ax.push(parseInt(Math.sqrt((p1[0]-p2[0])*(p1[0]-p2[0])+(p1[1]-p2[1])*(p1[1]-p2[1])+(p1[2]-p2[2])*(p1[2]-p2[2]))*property.srcR));
		var p1=camView_getPixel(property,0+0.001,0+0.001,property.db_dstw,property.db_dsth,true);
		var p2=camView_getPixel(property,property.db_dstw+0.001,0+0.001,property.db_dstw,property.db_dsth,true);
		ax.push(parseInt(Math.sqrt((p1[0]-p2[0])*(p1[0]-p2[0])+(p1[1]-p2[1])*(p1[1]-p2[1])+(p1[2]-p2[2])*(p1[2]-p2[2]))*property.srcR));

		var p1=camView_getPixel(property,property.db_dstw+0.001,0+0.001,property.db_dstw,property.db_dsth,true);
		var p2=camView_getPixel(property,property.db_dstw+0.001,property.db_dsth+0.001,property.db_dstw,property.db_dsth,true);
		ay.push(parseInt(Math.sqrt((p1[0]-p2[0])*(p1[0]-p2[0])+(p1[1]-p2[1])*(p1[1]-p2[1])+(p1[2]-p2[2])*(p1[2]-p2[2]))*property.srcR));
		var p1=camView_getPixel(property,property.db_dstw/2+0.001,0+0.001,property.db_dstw,property.db_dsth,true);
		var p2=camView_getPixel(property,property.db_dstw/2+0.001,property.db_dsth+0.001,property.db_dstw,property.db_dsth,true);
		ay.push(parseInt(Math.sqrt((p1[0]-p2[0])*(p1[0]-p2[0])+(p1[1]-p2[1])*(p1[1]-p2[1])+(p1[2]-p2[2])*(p1[2]-p2[2]))*property.srcR));
		var p1=camView_getPixel(property,0+0.001,0+0.001,property.db_dstw,property.db_dsth,true);
		var p2=camView_getPixel(property,0+0.001,property.db_dsth+0.001,property.db_dstw,property.db_dsth,true);
		ay.push(parseInt(Math.sqrt((p1[0]-p2[0])*(p1[0]-p2[0])+(p1[1]-p2[1])*(p1[1]-p2[1])+(p1[2]-p2[2])*(p1[2]-p2[2]))*property.srcR));

		var rx=Math.max(ax[0],ax[1],ax[2]);
		var ry=Math.max(ay[0],ay[1],ay[2]);

		if (property.aspect>=1) {
			ry=parseInt(rx/property.aspect);
		} else {
			rx=parseInt(ry*property.aspect);
		}

		if (rx<1) {rx=1;}
		if (ry<1) {ry=1;}

		return [rx,ry];
	}

	function callback(ok) {
		if (property.callback) {
			if (ok) {
				eval(property.callback);
			} else {
				eval(property.callbackerror);
			}
		}
	}
}
